import threading

from werkzeug.serving import make_server


class ServerThread(threading.Thread):
    def __init__(self, app):
        threading.Thread.__init__(self)
        self.srv = make_server("127.0.0.1", 5000, app)
        self.ctx = app.app_context()
        self.ctx.push()

    def run(self):
        # log.info("starting server")
        self.srv.serve_forever()

    def shutdown(self):
        self.srv.shutdown()



import importlib
import json
import os
import sys
from types import ModuleType

from cfn_tools import load_yaml
from flask import Flask, request

IGNORE_VARS = ["PATH", "LD_LIBRARY_PATH"]

app = Flask(__name__)
TEMPLATE_PATH = None


def start(port, template):
    global TEMPLATE_PATH  # pylint: disable=global-statement
    TEMPLATE_PATH = template
    app.run(port=port)


@app.route("/", methods=["POST"])
def local_function_api():
    return local_function(request.json)


def local_function(payload):
    set_up(payload)
    response = invoke(payload)
    tear_down(payload)
    return json.dumps(response)


def invoke(payload):
    if payload["code_uri"] == "__INLINE":
        function = load_inline(payload)
    else:
        function = load_code(payload)

    return function(payload["event"], None)


def load_code(payload):
    module = importlib.import_module(payload["file"])
    importlib.reload(module)
    return getattr(module, payload["function"])


def load_inline(payload):
    global TEMPLATE_PATH  # pylint: disable=global-statement
    template = load_yaml(open(TEMPLATE_PATH, "r").read())
    code = template["Resources"][payload["name"]]["Properties"]["InlineCode"]
    module = ModuleType("localfunk_index")
    exec(code, module.__dict__)  # pylint: disable=exec-used
    return getattr(module, payload["function"])


def set_up(payload):
    clear_env()
    set_env(payload.get("env"))
    if payload["code_uri"] != "__INLINE":
        sys.path.append(payload["code_uri"])


def tear_down(payload):
    unset_env(payload["env"])
    if payload["code_uri"] != "__INLINE":
        sys.path.remove(payload["code_uri"])


def clear_env():
    for key in os.environ.keys():  # pylint: disable=consider-iterating-dictionary
        os.environ.pop(key)


def set_env(env):
    for key, value in env.items():
        if key not in IGNORE_VARS:
            os.environ[key] = value


def unset_env(env):
    for key in env.keys():
        if key not in IGNORE_VARS:
            os.environ.pop(key)



version = "0.0.4"

name = "localfunk"
author = "Adam Baker"
author_email = "adambaker@keemail.me"
description = "Proxy Lambda functions to local code"
license = "MIT"
homepage = "https://github.com/atomraymaker/localfunk"



import argparse
import os
import subprocess

from pyngrok import ngrok

from localfunk import sam, server


def main():
    parser = parse()
    args = parser.parse_args()
    run(args)


def parse():
    parser = argparse.ArgumentParser(description="Proxy Lambdas to local code")
    parser.add_argument(
        "--template", type=str, default="template.yaml", help="path to template",
    )
    parser.add_argument(
        "--port", type=int, default=5000, help="port to run local server on",
    )
    parser.add_argument(
        "--url",
        type=str,
        default=None,
        help="url to proxy to, will skip starting ngrok",
    )
    parser.add_argument(
        "--guided",
        action="store_true",
        help="use guided in sam deploy, will use this by default if no samconfig.toml found",
    )
    parser.add_argument(
        "--server", action="store_true", help="only run local server",
    )
    return parser


def run(args):
    temp_template = "./locafunk_template.yaml"

    if args.server:
        server.start(args.port, args.template)
    else:
        # ngrok
        if args.url:
            public_url = args.url
        else:
            public_url = ngrok.connect(args.port)

        # build
        sam.build(public_url.split("//")[-1], args.template, temp_template)

        # deploy
        deploy_command = ["sam", "deploy", f"-t{temp_template}"]
        if args.guided or not os.path.exists("samconfig.toml"):
            deploy_command.append("--guided")
        subprocess.run(deploy_command, check=True)

        sam.cleanup(temp_template)

        # local server
        server.start(args.port, args.template)



import os

from cfn_tools import dump_yaml, load_yaml


def cleanup(template):
    if os.path.exists(template):
        os.remove(template)


def build(url, template_path, save_path):
    template_yaml = parse(template_path)
    functions = get_functions(template_yaml)
    update_template(template_yaml, functions, url)
    save_template(template_yaml, save_path)


def update_template(template_yaml, functions, url):
    for name, details in template_yaml["Resources"].items():
        if details["Type"] == "AWS::Serverless::Function":
            props = template_yaml["Resources"][name]["Properties"]
            if "CodeUri" in props:
                del props["CodeUri"]
            props["Handler"] = "index.handler"
            props["InlineCode"] = gen_code(functions[name], url)


def save_template(template_yaml, save_path):
    file = open(save_path, "w+")
    file.write(dump_yaml(template_yaml))
    file.close()


def gen_code(func, proxy):
    return f"""
import http.client
import json
import os

conn = http.client.HTTPConnection("{proxy}")
headers = {{"Content-type": "application/json"}}


def handler(event, context):
    data = {{
        "event": event, 
        "env": dict(os.environ.items()),
        "name": "{func['name']}",
        "code_uri": "{func['code_uri']}",
        "file": "{func['file']}",
        "function": "{func['function']}" 
    }}

    try:
        conn.request("POST", "/", json.dumps(data), headers)
        response = conn.getresponse()
        return json.loads(response.read().decode())
    except Exception as e:
        return {{ "error": str(e) }}
    """


def get_functions(template_yaml):
    functions = {}
    for name, details in template_yaml["Resources"].items():
        if details["Type"] == "AWS::Serverless::Function":
            props = details["Properties"]
            file, function = props["Handler"].split(".")

            if "CodeUri" in props:
                code_uri = props["CodeUri"]
            else:
                code_uri = "__INLINE"

            functions[name] = {
                "name": name,
                "code_uri": code_uri,
                "file": file,
                "function": function,
            }

    return functions


def parse(template_path):
    text = open(template_path, "r").read()
    return load_yaml(text)



import json
import os
from test.util.server import ServerThread

import flask

from localfunk import sam

app = flask.Flask("test_gen_code")


def test_build():
    path = ".localfunk.yaml"

    if os.path.exists(path):
        os.remove(path)

    sam.build("127.0.0.1:5000", "test/fixtures/template.yml", path)

    assert os.path.exists(path)

    if os.path.exists(path):
        os.remove(path)


def test_gen_code():
    server = ServerThread(app)
    server.start()

    path = "generated_code.py"
    if os.path.exists(path):
        os.remove(path)

    code = sam.gen_code(func("test_code"), "127.0.0.1:5000")
    file = open(path, "w+")
    file.write(code)
    file.close()

    import generated_code  # pylint: disable=import-error, import-outside-toplevel

    response = generated_code.handler({"event": "test"}, {"context": "test"})

    server.shutdown()

    assert response == {
        "event": {"event": "test"},
        "env": dict(os.environ.items()),
        "name": "Function",
        "code_uri": "app",
        "file": "test_code",
        "function": "test_code",
    }

    if os.path.exists(path):
        os.remove(path)


def test_gen_code_failure():
    server = ServerThread(app)
    server.start()

    path = "generated_code_fail.py"
    if os.path.exists(path):
        os.remove(path)

    code = sam.gen_code(func("test_code"), "127.0.0.1:5002")  # incorrect port
    file = open(path, "w+")
    file.write(code)
    file.close()

    import generated_code_fail  # pylint: disable=import-error, import-outside-toplevel

    response = generated_code_fail.handler({"event": "test"}, {"context": "test"})

    server.shutdown()

    assert response == {"error": "[Errno 61] Connection refused"}

    if os.path.exists(path):
        os.remove(path)


def test_get_functions():
    template_yaml = sam.parse(template_path="test/fixtures/template.yml")
    assert sam.get_functions(template_yaml=template_yaml) == {
        "FunctionOne": {
            "name": "FunctionOne",
            "code_uri": "app1",
            "file": "handler1",
            "function": "handle",
        },
        "FunctionTwo": {
            "name": "FunctionTwo",
            "code_uri": "__INLINE",
            "file": "index",
            "function": "handler",
        },
    }


def func(name):
    return {
        "name": "Function",
        "code_uri": "app",
        "file": name,
        "function": name,
    }


@app.route("/", methods=["POST"])
def hello():
    return json.dumps(flask.request.json)



from setuptools import find_packages, setup

import localfunk

setup(
    name=localfunk.__NAME__,
    version=localfunk.__VERSION__,
    author=localfunk.__AUTHOR__,
    author_email=localfunk.__AUTHOR_EMAIL__,
    description=localfunk.__DESCRIPTION__,
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    license=localfunk.__LICENSE__,
    url=localfunk.__HOMEPAGE__,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["flask", "pyngrok", "cfn-flip", "boto3"],
    entry_points={"console_scripts": ["localfunk=localfunk.main:main",]},
)



import json

from localfunk import server


def test_run_code():
    var_value = "test_run_code"
    payload = {
        "env": {"env_var": var_value},
        "event": {"var_name": "env_var"},
        "context": {},
        "name": "Function",
        "code_uri": "test/fixtures",
        "file": "function_code",
        "function": "run",
    }

    assert json.loads(server.local_function(payload)) == {"var_value": var_value}
